from eigenwerte_1 import *

""" Modifiziertes Signum, das 1 fuer Zahlen >= 0 gibt, -1 sonst
    Arguments:
    xx -- zu testendes Ding
"""
def my_sign(xx):
    return 1. if abs(xx) == 0 else xx/abs(xx)

""" Testest, ob Matrix auch Hessenbergform hat
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_hessenberg(AA, eps=1e-14):
    return np.max(np.abs(np.tril(AA, k=-2))) < eps

""" Testet, ob Matrix unitaer ist
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_unitary(AA, eps=1e-14):
    return np.max(np.abs(np.eye(np.shape(AA)[0])-AA.dot(np.conjugate(AA.T)))) < eps and np.max(np.abs(np.eye(np.shape(AA)[0])-np.conjugate(AA.T).dot(AA))) < eps

""" Testet ob QQ.T.dot(AA.dot(QQ)) == BB
    Arguments:
    AA -- Matrix
    BB -- Matrix
    QQ -- Matrix
    eps -- Genauigkeit
"""
def are_similar(AA, BB, QQ, eps=1e-14):
    return np.max(np.abs(np.conjugate(QQ.T).dot(AA.dot(QQ))-BB)) < eps

""" Testet, ob Matrix tridiagonal ist
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_tridiagonal(AA, eps=1e-14):
    return np.max(np.abs(np.tril(AA, k=-2))) < eps and np.max(np.abs(np.triu(AA, k=2))) < eps

""" Testet, ob Matrix bis auf eps 0 ist
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_zero(AA, eps=1e-14):
    return np.max(np.abs(AA)) < eps

""" Testtest, ob WW tatsaechlich die Eigenvektoren zu den Eigenwerten VV von AA sind
    Arguments:
    AA -- Matrix
    VV -- Eigenwerte
    WW -- Eigenvektoren
    eps -- Genauigkeit
"""
def test_eigenvalues(AA, VV, WW, eps):
    return np.max(np.abs(AA.dot(WW) - WW.dot(np.diag(VV))))

""" Transformiert eine Matrix AA mit Householder-Spiegelungen auf obere Hessenberg-Form
    Arguments:
    AA -- zu transformierende Matrix
    return_QQ -- ob Transformationsmatrix
    Rueckgabe
    HH -- transformierte Matrix in Hessenbergform
    QQ -- Transformationsmatrix sodass: QQ.T.conjugate().dot(AA.dot(QQ)) = HH
"""
def to_hessenberg(AA, return_QQ=False):
    if sp.issparse(AA):
        HH = AA.toarray()
    else:
        HH = AA.copy()
    # Transformationsmatrix
    QQ = np.eye(np.shape(AA)[0])
    if np.iscomplexobj(AA):
        QQ = QQ+0j
    for ii in xrange(0, np.shape(AA)[0]-2):
        # Spiegelungsvektor konstruieren
                                  
                                   
                  
                                               
                                                 
        # Transformationsmatrix update, auf effizienten Anwendung von np.outer(vv, vv) achten!
                                                                          
        # Spiegelung von links anwenden
                                                                 
                                                                                  
        # Spiegelung von rechts anwenden
                                                                          
    # Rueckgabe
    if return_QQ:
        return   ,                 
    else:
        return   

""" Gibt 2x2 Givens Matrix, die JJ = AA[jj,kk], mit Hilfe von II = AA[ii, kk] eliminiert,
    also die (verkleinerte) Givens-Matrix G(ii, jj)
    Arguments:
    II -- Wert mit dem eliminiert werden soll
    JJ -- Wert der eliminiert werden soll
"""
def get_givens(II, JJ):
    # Cosinus und Sinus berechnen
                      
                                                                 
                            
                                                
         
              
              
    # Rueckgabe als 2x2 Matrix
    return np.array([[cc, ss], [-ss.conjugate(), cc]])

""" Modifizierte QR-Zerlegung fuer Matrizen in Hessenbergform mit Givens-Rotationen
    Arguments:
    HH -- Matrix in Hessenbergform, die in QQ.dot(RR) == HH zerlegt wird
    only_SS -- nur RR.dot(QQ) zurueckgeben
    Rueckgabe
    QQ -- Transformation
    RR -- rechte Seite
    SS -- RR.dot(QQ)
"""
def hessenberg_qr(HH, only_SS=False):
    RR = HH.copy()
    # Platz zum Speichern der Rotationen
    if np.iscomplexobj(RR):
        rotations = np.array([np.eye(2)+0j for ii in xrange(np.shape(HH)[0]-1)])
    else:
        rotations = np.array([np.eye(2) for ii in xrange(np.shape(HH)[0]-1)])
    # Rotationen von links anwenden, auf effiziente Anwendung achten, nur 2 Zeilen aendern!
    for ii in                             :
        # Rotation berechnen und speichern
        rotations[ii] =                                  
        # Rotation von links anwenden
                                                            
    SS = RR.copy()
    # Platz fuer QQ
    if np.iscomplexobj(RR):
        QQ = np.eye(np.shape(HH)[0])+0j
    else:
        QQ = np.eye(np.shape(HH)[0])
    # Rotationen von rechts anwenden, auf effizienten Anwendung achten, nur 2 Spalten aendern
    for ii in                             :
                                                                                
                                                                                

    # Rueckgabe
    if only_SS:
        return SS
    else:
        return QQ, RR, SS

""" QR-Verfahren mit Shift
    Arguments:
    AA -- Matrix
    eps -- gewuenschte Genauigkeit, maximaler Betrag von Subdiagonalelementen
    Rueckgabe:
    VV -- Eigenwerte
    WW -- Eigenvektoren
    iterations -- Anzahl Iterationen
"""
def qr_shift(AA, eps=1e-10):
    # Hessenbergtransformation
                                    
    # Schleife
    err = np.max(np.abs(np.diag(HH,-1)))
    iterations = 0
    while err > eps*np.max(np.abs(HH)):
        iterations += 1
        # Shift 
                                              
        # QR Zerlegung
                                         
                       
        # Shift Teil 2
                
        # Fehler
        err = np.max(np.abs(np.diag(HH,-1)))
    # Rueckgabe
    return            ,   , iterations

""" Malt geforderte Bilder mit Einschliessungen und Eigenwerten
"""
def plot_eigenvalue_stuff(AA, eps=1e-10, fig=None):
    if fig == None:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    # Einschliessungen malen
                            
                          
    ax.set_xlabel(r'$\mathbb{R}$')
    ax.set_ylabel(r'$\mathrm{i} \mathbb{R}$')
    # Eigenwerte berechnen
    VV, WW, iterations = qr_shift(AA, eps)
    # Eigenwerte malen
    ax.plot(           ,            , 'kx')
    # zeigen
    fig.show()
    # Rueckgabe
    return VV, WW, iterations

""" Testfaelle und Aufgaben
"""
if __name__ == '__main__':
    eps = 1e-15

    print "\nGivens Tests"
    aa = 1; bb = 2
    GG = get_givens(aa, bb)
    print "    ", is_unitary(GG), is_zero(GG.dot(np.array([aa, bb]))[1])
    aa = 0; bb = 1
    GG = get_givens(aa, bb)
    print "    ", is_unitary(GG), is_zero(GG.dot(np.array([aa, bb]))[1])
    aa = 1+1j; bb = 1+2j
    GG = get_givens(aa, bb)
    print "    ", is_unitary(GG), is_zero(GG.dot(np.array([aa, bb]))[1])

    print "\nQR, Tests"

    print "\n  2D real"
    aa, bb = rand(2); nn = np.sqrt(aa*aa+bb*bb)
    aa /= nn; bb /= nn
    GG = np.array([[aa, bb], [-bb, aa]])
    print "    ", is_unitary(GG)
    AA = GG.T.dot(np.diag(rand(2)).dot(GG))
    HH, QQ = to_hessenberg(AA, True)
    print "    ", is_hessenberg(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
    QQ, RR, SS = hessenberg_qr(HH)
    print "    ", is_hessenberg(RR), is_unitary(QQ), is_zero(QQ.dot(RR)-HH) 
    VV, WW, iterations = plot_eigenvalue_stuff(AA, eps)
    print "    ", iterations, " iterations"
    print "    ", test_eigenvalues(AA, VV, WW, eps), "error"

    print "\n  3D real"
    AA = np.array([[4., 0., 3.], [0., -1., 1.], [1., 1., 0.]])
    HH, QQ = to_hessenberg(AA, True)
    print "Test 0: ", is_hessenberg(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
    QQ, RR, SS = hessenberg_qr(HH)
    print "    ", is_hessenberg(RR), is_unitary(QQ), is_zero(QQ.dot(RR)-HH)
    VV, WW, iterations = plot_eigenvalue_stuff(AA, eps)
    print "    ", iterations, " iterations"
    # nicht diagonalisierbar, Eigenwerttest bringt hier nix...
    plt.gcf().savefig('test_0.pdf')

    print "\n  10D real"
    AA = rand(10,10); AA+=AA.T
    HH, QQ = to_hessenberg(AA, True)
    print "    ", is_hessenberg(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
    QQ, RR, SS = hessenberg_qr(HH)
    print "    ", is_hessenberg(RR), is_unitary(QQ), is_zero(QQ.dot(RR)-HH) 
    VV, WW, iterations = plot_eigenvalue_stuff(AA, eps)
    print "    ", iterations, " iterations"
    print "    ", test_eigenvalues(AA, VV, WW, eps), "error"

    print "\n  10D complex"
    AA = rand(10,10)+rand(10,10)*1j; AA += AA.T.conjugate()
    HH, QQ = to_hessenberg(AA, True)
    print "    ", is_tridiagonal(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
    QQ, RR, SS = hessenberg_qr(HH)
    print "    ", is_hessenberg(RR), is_unitary(QQ), is_zero(QQ.dot(RR)-HH) 
    VV, WW, iterations = plot_eigenvalue_stuff(AA, eps)
    print "    ", iterations, " iterations"
    print "    ", test_eigenvalues(AA, VV, WW, eps), "error"

    # Aufgabe
    print "\nAufgabe"
    for ii in xrange(10, 21, 10):
        print "\n  ", ii
        AA = get_gram(ii)
        HH, QQ = to_hessenberg(AA, True)
        print "    ", is_tridiagonal(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
        QQ, RR, SS = hessenberg_qr(HH)
        print "    ", is_hessenberg(RR), is_unitary(QQ), is_zero(QQ.dot(RR)-HH) 
        VV, WW, iterations = plot_eigenvalue_stuff(AA, eps)
        print "    ", iterations, " iterations"
        print "    ", test_eigenvalues(AA, VV, WW, eps), "error"
        plt.gcf().savefig(str(ii)+'_eigenwerte.pdf')
